# Licensed under the EUPL.

from arcface.lib.models import ArcFaceModel
import arcface
from arcface.lib.utils import l2_norm
import tensorflow as tf
import numpy as np
import cv2
import os
import requests


class ArcFace():
    def __init__(self): 
        data_path = os.path.join(os.path.dirname(arcface.__file__), "data")
        tflite_path = os.path.join(data_path, "model.tflite")
        if not os.path.isdir(data_path):
            os.mkdir(data_path)
        if not os.path.exists(tflite_path):
            self._download_model_file(tflite_path)
    
        self.interpreter = tf.lite.Interpreter(model_path=tflite_path)
        self.interpreter.allocate_tensors()
        self.input_details = self.interpreter.get_input_details()
        self.output_details = self.interpreter.get_output_details()

    def calc_emb(self, imgs):
        """Calculates the embedding from an (array of) images. These images can be cv2-image-files or a path to a file.
        Parameters:
            imgs (str|list): either a list of images or the image itself. The image can be a cv2-image or a path to a file. The images should already be aligned!
        Returns:
            ndarray: 512-d embedding of the supplied image(s)
        Example:
            calc_emb("~/Downloads/test.jpg")
        """
        if isinstance(imgs, list):
            return self._calc_emb_list(imgs)
        return self._calc_emb_single(imgs)


    def get_distance_embeddings(self, emb1, emb2):
        """Calculates the distance (L2 norm) between two embeddings. Larger values imply more confidence that the two embeddings are from different people.
        Parameters:
            emb1 (ndarray): embedding of a person (e.g. generated by calc_emb(...))
            emb2 (ndarray): embedding of a person (e.g. generated by calc_emb(...))
        Returns:
            int: distance between emb1 and emb2
        Example:
            get_distance_embeddings(calc_emb("person1.jpg"), calc_emb("person2.jpg"))
        """
        diff = np.subtract(emb1, emb2)
        dist = np.sum(np.square(diff))
        return dist
        
    def _download_model_file(self, path):
        print("Downloading pre-trained Tensorflow Light model (161 MB)...")
        eflite_path = os.path.join(os.path.dirname(arcface.__file__), "data/model.tflite")
        tflight_model = requests.get("https://cloud.ins.jku.at/index.php/s/g2YDT8Zn9RkzsEy/download")
        open(path, 'wb').write(tflight_model.content)

    def _calc_emb_list(self, imgs):
        embs = []
        for img in imgs:
            embs.append(self._calc_emb_single(img))
        return embs

    def _calc_emb_single(self, img):
        if isinstance(img, str):
            img = cv2.imread(img)
            if img is None:
                raise FileNotFoundError(errno.ENOENT, os.strerror(errno.ENOENT), img)
        
        img = cv2.cvtColor(img, cv2.COLOR_BGR2RGB)
        img = cv2.resize(img, (112, 112))
        img = img.astype(np.float32) / 255.
        if len(img.shape) == 3:
            img = np.expand_dims(img, 0)
        self.interpreter.set_tensor(self.input_details[0]['index'], img)
        self.interpreter.invoke()
        output_data = self.interpreter.get_tensor(self.output_details[0]['index'])

        emb = l2_norm(output_data)
        return emb[0]
